from typing import Any, Dict

from httpx import AsyncClient


def create_client(base_url: str, headers: Dict[Any, Any]) -> AsyncClient:
    """Create AsyncClient."""
    return AsyncClient(base_url=base_url, headers=headers)


def prepare_headers(token: str) -> Dict:
    """Prepare headers for client."""
    if token:
        return {"Authorization": f"token {token}"}
    return {}



class InvalidPathException(Exception):
    """No such file or directory."""


class NotebookDoesNotExist(Exception):
    """Notebook file does not exist."""


class SessionExists(Exception):
    """Session name already in use."""


class NBFormatModuleNotFound(Exception):
    """NBFormat module not found."""



from typing import Dict

from httpx import AsyncClient, codes

from nbr.exceptions import InvalidPathException


async def get_contents(path: str, client: AsyncClient) -> Dict:
    """Get content by path."""
    url = f"/contents/{path}"

    response = await client.get(url)

    if response.status_code == codes.NOT_FOUND.value:
        raise InvalidPathException(f"No such file or directory: {path}")

    return response.json()



from enum import Enum
from types import TracebackType
from typing import Any, Awaitable, Optional, Type, TypeVar

from httpx import AsyncClient

from nbr.api import JupyterAPI
from nbr.kernel import Kernel
from nbr.notebook import Notebook
from nbr.schemas.result import RunResult
from nbr.schemas.session import CreateSession, Session
from nbr.utils.client import create_client, prepare_headers
from nbr.utils.session import create_session, delete_session

TNotebookRunner = TypeVar("TNotebookRunner", bound="NotebookRunner")


class RunnerState(Enum):
    UNOPENED = 1
    OPENED = 2
    CLOSED = 3


class NotebookRunner:
    def __init__(
        self,
        *,
        notebook: Notebook,
        on_notebook_start: Optional[Awaitable[Any]] = None,
        on_notebook_end: Optional[Awaitable[Any]] = None,
        jupyter_api: JupyterAPI = JupyterAPI(),
    ) -> None:
        self._state: RunnerState = RunnerState.UNOPENED

        self.notebook: Notebook = notebook

        self.on_notebook_start = on_notebook_start
        self.on_notebook_finish = on_notebook_end

        self.jupyter_api = jupyter_api

        self._client: AsyncClient = create_client(
            base_url=f"http://{self.jupyter_api.host}:{self.jupyter_api.port}/api",
            headers=prepare_headers(self.jupyter_api.token),
        )

        self._session: Session
        self._kernel: Kernel

    async def execute_all_cells(self) -> RunResult:
        if self._state != RunnerState.OPENED:
            raise RuntimeError("Create NotebookRunner instance first.")

        if self.on_notebook_start:
            await self.on_notebook_start

        run_result = await self._kernel.execute(cells=self.notebook.cells)
        self.notebook.cells = run_result.cells

        if self.on_notebook_finish:
            await self.on_notebook_finish

        return run_result

    async def __aenter__(self: TNotebookRunner) -> TNotebookRunner:
        if self._state != RunnerState.UNOPENED:
            raise RuntimeError(
                "Cannot create a NotebookRunner instance more than once.",
            )

        self._state = RunnerState.OPENED
        self._session = await create_session(
            session_data=CreateSession(
                name=self.notebook.name, path=self.notebook.path
            ),
            client=self._client,
        )
        self._kernel = Kernel(session=self._session)
        await self._kernel.start(
            base_url=f"{self.jupyter_api.host}:{self.jupyter_api.port}/api"
        )

        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        await delete_session(session_id=self._session.id, client=self._client)

        self._state = RunnerState.CLOSED



from typing import List

from httpx import AsyncClient

from nbr.schemas.session import CreateSession, Session


async def get_sessions(*, client: AsyncClient) -> List[Session]:
    """Get all sessions."""
    response = await client.get("/sessions")
    all_sessions = [Session(**data) for data in response.json()]

    return all_sessions


async def create_session(
    *, session_data: CreateSession, client: AsyncClient
) -> Session:
    """Create a new session."""
    response = await client.post("/sessions", json=session_data.dict())
    return Session(**response.json())


async def delete_session(*, session_id: str, client: AsyncClient) -> None:
    """Delete session by id."""
    url = f"/sessions/{session_id}"
    await client.delete(url)



import asyncio
import json
from typing import List

import nbformat
from websockets.legacy.client import WebSocketClientProtocol

from nbr.schemas.message import Content
from nbr.schemas.result import ExecutionStatus, RunResult
from nbr.schemas.session import Session
from nbr.utils.message import create_message
from nbr.utils.websocket import connect_websocket


class Kernel:
    def __init__(self, *, session: Session) -> None:
        self._session: Session = session
        self._channel_tasks: List[asyncio.Task] = []
        self._status: ExecutionStatus = ExecutionStatus.WAITING

        self._websocket: WebSocketClientProtocol

        self._cells: List
        self._current_cell: int

    async def listen_server(self) -> None:
        """Listen server messages."""
        while True:
            msg = await self._websocket.recv()

            msg_json = json.loads(msg)
            content = msg_json["content"]

            if "data" in content:
                content["output_type"] = "execute_result"
                self._cells[self._current_cell].outputs = [
                    nbformat.NotebookNode(content)
                ]

            if "status" in content:
                self._cells[self._current_cell]["execution_count"] = (
                    self._current_cell + 1
                )
                self._current_cell += 1

            if (
                "status" in content
                and "execution_count" in content
                and content["execution_count"] == len(self._cells)
            ):
                self._status = ExecutionStatus.SUCCESS

                await self._stop()
                break

            if "status" in content and content["status"] == "aborted":
                self._status = ExecutionStatus.ERROR

                await self._stop()
                break

    async def start(self, base_url: str) -> None:
        self._websocket = await connect_websocket(
            base_url=base_url, session=self._session
        )
        self._channel_tasks.append(asyncio.create_task(self.listen_server()))

    async def _stop(self) -> None:
        self._channel_tasks[-1].cancel()
        await self._websocket.close()

    async def execute(self, cells: List[nbformat.NotebookNode]) -> RunResult:
        self._cells = cells
        self._current_cell = 0

        for cell in cells:
            code = cell["source"]

            content = Content(code=code)
            message = create_message(
                channel="shell",
                msg_type="execute_request",
                session=self._session.name,
                content=content,
            )

            await self._websocket.send(message)

        await asyncio.gather(*self._channel_tasks)

        return RunResult(status=self._status, cells=self._cells)



from datetime import datetime
from typing import Union
from uuid import uuid4

from pydantic import BaseModel


class Metadata(BaseModel):
    cellId: str = uuid4().hex
    deletedCells: list = []
    recordTiming: bool = False


class Content(BaseModel):
    allow_stdin: bool = False
    code: str = ""
    silent: bool = False
    stop_on_error: bool = True
    store_history: bool = True
    user_expressions: dict = {}


class Header(BaseModel):
    date: str = str(datetime.utcnow())
    msg_id: str = uuid4().hex
    msg_type: str
    session: str
    username: str = ""
    version: str = "5.2"


class Message(BaseModel):
    buffers: list = []
    channel: str
    content: Union[Content, dict]
    header: Header
    metadata: Union[Metadata, dict]
    parent_header: dict = {}



from typing import Optional, Union

from nbr.schemas.message import Content, Header, Message, Metadata


def create_message(
    channel: str,
    msg_type: str,
    session: str,
    content: Optional[Union[Content, dict]] = None,
    metadata: Optional[Union[Metadata, dict]] = None,
) -> str:
    """Generate a message using a template."""
    if not content:
        content = {}

    if not metadata:
        metadata = {}

    header = Header(msg_type=msg_type, session=session)

    message_data = {
        "channel": channel,
        "header": header,
        "content": content,
        "metadata": metadata,
    }

    message = Message(**message_data)

    return message.json()



from enum import Enum
from typing import List

import nbformat
from pydantic import BaseModel


class ExecutionStatus(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WAITING = "waiting"


class RunResult(BaseModel):
    status: ExecutionStatus
    cells: List[nbformat.NotebookNode]



from websockets.legacy.client import Connect, connect

from nbr.schemas.session import Session


def connect_websocket(base_url: str, session: Session) -> Connect:
    """Connect to websocket."""

    kernel_id = session.kernel.id
    session_id = session.id
    url = f"ws://{base_url}/kernels/{kernel_id}/channels?session_id={session_id}"

    return connect(url)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbr', 'nbr.schemas', 'nbr.utils']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.22.0,<0.23.0',
 'nbformat>=5.3.0,<6.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'websockets>=10.2,<11.0']

setup_kwargs = {
    'name': 'nbr',
    'version': '1.0.0b0',
    'description': 'Jupyter notebooks runner',
    'long_description': '[![CI](https://github.com/zhivykh/nbr/workflows/CI/badge.svg)](https://github.com/zhivykh/nbr/actions/workflows/main.yml)\n[![codecov](https://codecov.io/gh/zhivykh/nbr/branch/main/graph/badge.svg?token=8BQOVVCL6B)](https://codecov.io/gh/zhivykh/nbr)\n[![Python 3.10](https://img.shields.io/badge/python-3.10-blue.svg)](https://www.python.org/downloads/release/python-3100/)\n[![Stable Version](https://img.shields.io/pypi/v/nbr?color=blue)](https://pypi.org/project/nbr/)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n\n# nbr\nNBR lets you **run** local and remote jupyter-notebooks.\n\n## Installation\nIn a terminal, run:\n```\npip install nbr\n```\n\n## Usage\n\nLaunch a Jupyter server:\n```\njupyter server\n```\n\nExecution a local notebook, using a remote server:\n\n\n```python\nimport asyncio\nfrom nbr import NotebookRunner, Notebook, JupyterAPI, ExecutionStatus\n\n\nasync def main() -> None:\n    jupyter_api = JupyterAPI()\n    notebook = Notebook.read_file(path="Untitled.ipynb")\n\n    async with NotebookRunner(notebook=notebook, jupyter_api=jupyter_api) as runner:\n        result = await runner.execute_all_cells()\n\n        if result.status == ExecutionStatus.SUCCESS:\n            notebook.save(path="Executed.ipynb")\n    \nif __name__ == "__main__":\n    asyncio.run(main())\n```',
    'author': 'Nick Zhivykh',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zhivykh/nbr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, List

import nbformat

from nbr.api import JupyterAPI
from nbr.utils.client import create_client, prepare_headers
from nbr.utils.contents import get_contents


class Notebook:
    def __init__(self, *, name: str, path: str) -> None:
        self.name = name
        self.path = path
        self.cells: List[nbformat.NotebookNode] = []

        self._remote: bool = False

    @property
    def is_remote(self) -> bool:
        return self._remote

    def save(self, path: str) -> None:
        if not self._remote:
            nbformat.write(nbformat.NotebookNode(self.dict()), path)

    @classmethod
    async def read_remote(
        cls, path: str, jupyter_api: JupyterAPI = JupyterAPI()
    ) -> "Notebook":
        client = create_client(
            base_url=f"http://{jupyter_api.host}:{jupyter_api.host}/api",
            headers=prepare_headers(jupyter_api.token),
        )
        notebook_name = path.split("/")[-1]
        notebook = cls(name=notebook_name, path=path)

        nb_content = await get_contents(path=path, client=client)
        notebook.cells = nb_content["content"]["cells"]

        notebook._remote = True

        return notebook

    @classmethod
    def read_file(cls, path: str) -> "Notebook":
        nb_file = nbformat.read(path, as_version=4)
        notebook_name = path.split("/")[-1]

        notebook = cls(name=notebook_name, path=path)
        notebook.cells = nb_file.cells

        return notebook

    def dict(self) -> Dict:
        return {
            "cells": self.cells,
            "metadata": {"language_info": {"name": "python"}},
            "nbformat": 4,
            "nbformat_minor": 2,
        }



from typing import Optional

from pydantic import BaseModel


class Kernel(BaseModel):
    """Kernel model."""

    id: str
    name: str
    last_activity: str
    execution_state: str
    connections: int


class Notebook(BaseModel):
    """Notebook model."""

    path: Optional[str] = None
    name: str


class Session(BaseModel):
    """Session model."""

    id: str
    path: str
    name: str
    type: str
    kernel: Kernel
    notebook: Notebook


class KernelName(BaseModel):
    """Kernel name scheme."""

    name: str = "python3"


class CreateSession(BaseModel):
    """Session scheme."""

    kernel: KernelName = KernelName()
    name: str
    path: str
    type: str = "notebook"
